//********************************************************************
// File EbFile.h              A simple file-access class based on IO.h
//********************************************************************
// Copyright @2013 by Stan Sykora and Extra Byte.
#ifndef _EbFile_h_
#define _EbFile_h_

// Defaults:
// All File and Directory names, as well as all Path strings are
// now in wide characters (wchar_t, LPWSTR), while actual file
// contents are assumed to be ASCI.

#include "EbApiFile.h"

// This class is a generic file manager. It is fully portable
// (OS-independent) but it relies on the OS encapsulation of
// file functions in EbFileApi.

#define DEFAULT_SDIGITS	 15  // Significant digits of double precision

// cEbFile fFlags:
// -------------------------------------------------------------------
	// Permanent flags
#define File_ALIAS		0x80000000
#define File_DISCRIMINATOR 	0x40000000
#define File_SAFEWRITE 		0x20000000
#define File_UNICODE 		0x00080000
#define File_ASCI 		0x00040000
#define File_UNIX 		0x00020000
	// Volatile flags (change upon Open/Close)
#define File_NOWRITE 		0x00000001
#define File_VOLATILEFLAGS	0x0000FFFF

// fErrorCode:
// -------------------------------------------------------------------
#define FERR_DiskNotAvailable (FERR_DiskAbsent|FERR_DiskNotDos \
  |FERR_DiskNotReady|FERR_WrongDisk|FERR_DiskFull|FERR_WriteProtected)

// Scan File Section flags
// -------------------------------------------------------------------
#define SFS_USERBREAK	1
#define SFS_NODATA	2
#define SFS_NOSECTION	4
#define SFS_NOFILE	8
#define SFS_ALL SFS_USERBREAK|SFS_NOSECTION|SFS_NODATA|SFS_NOFILE

//--------------------------------------------------------------------
class cEbFile
// -------------------------------------------------------------------
{
public:
  DWORD  Tag;				      // Reserved for the User
protected:
  LPWSTR fName;        				          // File name
  HANDLE fHandle;   		    // OS identifier of an opened file
  DWORD  fFlags;                        // See the cEbFile flags above
  DWORD  fError;         	      // OS-dependent error identifier
  union {                                    // Associated file buffer
    LPSTR   fBuffer;
    LPWSTR  fBufferW;
  };
private:
  cInt64 fPointer;    // An auxiliary related to the OS's file pointer
  HANDLE fFindHandle;
public:
  virtual ~cEbFile();
  cEbFile();
  cEbFile(LPCWSTR name,LPCWSTR dir=NULL,LPCWSTR ext=NULL);
  cEbFile(LPCSTR name);
  void    MakeAlias(HANDLE fh,DWORD flags=fRead,
  		    LPCWSTR fullname=NULL,DWORD noreset=0);

  inline  DWORD Error(){return fError;}
                                            // These operate on fName:
  inline  LPWSTR FullName(){return fName;}
  LPWSTR  FullName(LPWSTR* pDest);
  LPWSTR  FullApiName(LPWSTR* pDest=NULL);
  LPWSTR  CleanName(LPWSTR* pDest=NULL);
  LPWSTR  Extension(LPWSTR* pDest=NULL);
  LPWSTR  Path(LPWSTR* pDest=NULL);
  LPWSTR  DiskRootPath(LPWSTR* pDest=NULL);

                                      // These create ex-novo strings:
  LPWSTR  CurrentDirectoryPath(LPWSTR* pDest=NULL);
  RETOK   CreateFileDirectory();

  inline  HANDLE Handle()      {return fHandle;}
  inline  DWORD  Flags()       {return fFlags;}
  inline  LPSTR  Buffer()      {return fBuffer;}
  inline  LPWSTR BufferW()     {return fBufferW;}

  inline  RETOK  isReadOnly()  {return fFlags & File_NOWRITE;}
  inline  RETOK  isOpen()      {return isHandleValid(fHandle);}
  inline  RETOK  isNotOpen()   {return isHandleInvalid(fHandle);}
  inline  RETOK  isAlias()     {return fFlags & File_ALIAS;}
  inline  RETOK  isFindActive(){return isHandleValid(fFindHandle);}

  inline  RETOK  isSafeWrite() {return fFlags & File_SAFEWRITE;}
  inline  void   SetSafeWrite(){fFlags |= File_SAFEWRITE;}

  LONG    isPresent();
  RETERR  MediaStatus();
  RETOK   FindFirst(LPCWSTR wildname);
  RETOK   FindNext();
  RETOK   FindEnd();

  virtual RETOK Reset(DWORD keepname = 0);
  RETOK   FlushFBuffers();
  RETOK   Kill();
  RETOK   KillIfEmpty();
  virtual RETOK KillIfBlank(){return Kill();}

  LONG    isNotEmpty();
  virtual LONG isNotBlank(){return isNotEmpty();}
  virtual LPCSTR Discriminator(){return NULL;}

  RETOK	  SetName(LPCWSTR name,LPCWSTR path=NULL,LPCWSTR ext=NULL);
  RETOK	  SetName(LPCSTR name);
  RETOK   AdaptedNameAndPath
	     (LPCWSTR newName,LPCWSTR defaultName,LPCWSTR ext);
  RETOK   ChangeExtension(LPCWSTR newext);

  RETOK   TryToSelectName(LPCWSTR title,
  	      LPCWSTR inidir,LPCWSTR defext,LPCWSTR filter,
              DWORD flags,HANDLE ownerwin=NULL,DWORD* retflags=NULL);

  RETOK   SelectName(LPCWSTR title,
  	      LPCWSTR inidir,LPCWSTR defext,LPCWSTR filter,
              DWORD flags,HANDLE ownerwin=NULL,DWORD* retflags=NULL);

  virtual RETOK Close(DWORD keepbuffer=0);
  RETOK   CloseAlias(DWORD keepbuffer=0);

  virtual RETOK Open(DWORD flags=fDefaults);
  RETOK   OpenNew(LPCSTR msgTitle="",DWORD share=0);
  RETOK   OpenOrCreate(LPCSTR msgTitle="",DWORD ini=fRewind);
  RETOK	  OpenExistingForWrite(LPCSTR msgTitle="",DWORD ini=fRewind);
  RETOK	  OpenForRead(LPCSTR msgTitle="");

  cInt64  Pointer();
  inline  RETOK RefreshPointer(){return MoveBy((DWORD)0);}
  RETOK   MoveTo(cInt64 newPtr);
  RETOK   MoveBy(cInt64 delta);
  RETOK   MoveToEnd();
  inline  RETOK MoveToStart(){return MoveTo((DWORD)0);}
  virtual RETOK Rewind()     {return MoveToStart();}

  RETOK   GetSize(cInt64* size);
  RETOK   TruncateAtCurrentPosition();

  LONG	  LocateString (LPCSTR str,
  		        fSEARCH flags=sDefault,DWORD strlen=0);
  LONG	  LocateStringEx(LPCSTR str,
  		        fSEARCH flags=sDefault,DWORD strlen=0);

                            		    // Read.... Return values:
  LONG	  Read(void* mem,DWORD bytes=1);    // >0 number of chars read
                                            //  0 end-of-file
                                            // <0 disk access error
  LONG    ReadStr(LPSTR str,DWORD strlen = 0,
  			LPCSTR locator=NULL,fSEARCH flags=sDefault);
  LONG	  ReadLine(LPSTR* pline=NULL,
  			LPCSTR locator=NULL,fSEARCH flags=sDefault);
  LONG	  ReadDword(DWORD* dw,
  	    		LPCSTR locator=NULL,fSEARCH flags=sDefault,
                        LPSTR* pptrailer=NULL,DWORD radix=10);
  LONG	  ReadLong(LONG* l,
  			LPCSTR locator=NULL,fSEARCH flags=sDefault,
  			LPSTR* pptrailer=NULL);
  LONG	  ReadDouble(double* l,LPCSTR locator=NULL,
  			fSEARCH flags=sDefault,LPSTR* pptrailer=NULL);

  LONG	  ByteIterator(DWORD(*Do)(BYTE byte,DWORD udata),DWORD data);
  LONG	  LineIterator(DWORD(*Do)(LPSTR str,DWORD usrdat),DWORD data);

  RETOK	  Write(LPCVOID mem,DWORD bytes);
  RETOK	  WriteStr (LPCSTR str);
  RETOK	  WriteStr (LPCSTR str1,LPCSTR str2,
                    LPCSTR str3=NULL,LPCSTR str4=NULL,LPCSTR str5=NULL);
  RETOK	  WriteLine();
  RETOK	  WriteLine(LPCSTR str);
  RETOK	  WriteLine(LPCSTR str1,LPCSTR str2,
                    LPCSTR str3=NULL,LPCSTR str4=NULL);
  RETOK	  WriteLineW(LPCWSTR str1,LPCWSTR str2,
                    LPCWSTR str3=NULL,LPCWSTR str4=NULL);
  RETOK	  WriteDword(DWORD dw,
                    LPCSTR lead1=NULL,LPCSTR lead2=NULL,
  	            LPCSTR trail=NULL,DWORD radix=10);
  RETOK	  WriteDwordLine(DWORD dw,
                    LPCSTR lead1=NULL,LPCSTR lead2=NULL,
  	            LPCSTR trail=NULL,DWORD radix=10);
  RETOK	  WriteLong(LONG l,
  	    LPCSTR lead1=NULL,LPCSTR lead2=NULL,LPCSTR trail=NULL);
  RETOK	  WriteLongLine(LONG l,
  	    LPCSTR lead1=NULL,LPCSTR lead2=NULL,LPCSTR trail=NULL);
  RETOK	  WriteDouble(double f,LPCSTR lead1=NULL,
  	    LPCSTR lead2=NULL,LPCSTR trail=NULL,
            DWORD precision = DEFAULT_SDIGITS);
  RETOK	  WriteDoubleLine(double f,LPCSTR lead1=NULL,
  	    LPCSTR lead2=NULL,LPCSTR trail=NULL,
            DWORD precision = DEFAULT_SDIGITS);

  RETOK   DeleteSection(cInt64 from,cInt64 size);
  RETOK   InsertSection(cInt64 at,void* newdata,DWORD size);
  RETOK   ReplaceSection
            (cInt64 from,cInt64 size,void* newdata,DWORD newsize);

  RETERR  ScanFileSectionLines( LPCSTR sectionStart,
                                LPCSTR sectionEnd,
                                DWORD  groomFlags,
                                DWORD  errsToShow,
                                DWORD  (*Do)(void*,LPSTR),
                                void*  UserData,
                                char   commentChr=0,
                                LPCSTR msgTitle="");

  LONG    OpenWithDiscriminator(LPCSTR discr,LPCSTR msgTitle="");
protected:
  inline  RETOK Ok() {fError = 0;return 1;}
  RETOK   Err();
  virtual void UserMessage(LPCSTR msgTitle,LPCSTR text);
private:
  LPSTR   ReturnString(LPSTR* dest,LPSTR str,DWORD copy);
  LPWSTR  ReturnStringW(LPWSTR* dest,LPWSTR str,DWORD copy);
  LONG	  PlainLineRead(LPSTR* pline);
};

class cEbFileWithDiscriminator
// -------------------------------------------------------------------
// cFileWithDiscriminator's are ASCI files whose
// first line starts with a predefined string called discriminator.
:public cEbFile
{
public:
protected:
private:
  LPCSTR  fDiscriminator;		   // The discriminator string
public:
  virtual ~cEbFileWithDiscriminator();
  cEbFileWithDiscriminator(LPCSTR discr = "");
  LPCSTR  Discriminator(){return fDiscriminator;}
  RETOK	  SetDiscriminator(LPCSTR discr);

  RETOK   Rewind();
  LONG    isNotBlank();
  RETOK   KillIfBlank();

  RETOK Open(DWORD flags=fDefaults);

protected:
private:
};

//--------------------------------------------------------------------
// Auxiliary functions
// -------------------------------------------------------------------

RETERR KillEbFile(cEbFile** pfile);
RETERR KillEbFileWithDiscriminator(cEbFileWithDiscriminator** pfile);

//--------------------------------------------------------------------
#endif



