 // *******************************************************************
// File EbReal.h                         Functions based on REAL types
// *******************************************************************
// Copyright @2014 by Stan Sykora and Extra Byte.
#ifndef _EbReal_h_
#define _EbReal_h_

#include "EbTypedefs.h"
#include "EbTemplates.h"
#include <math.h>
#include <stdlib.h>

// ===================================================================
// REAL macros
// ===================================================================

// Floating point handling (double only, little endian):
#define isFpNAN(fptr)         (Word(3,(fptr))==0xffff)
#define isFpPlusINF(fptr)     ((!Dword(0,(fptr)))&&(Dword(1,(fptr))==0x7ff00000))
#define isFpMinusINF(fptr)    ((!Dword(0,(fptr)))&&(Dword(1,(fptr))==0xfff00000))
#define isFpZero(fptr)        (!(Dword(0,(fptr))||Dword(1,(fptr))))

#define isFpNegative(fptr)    ((Dword(1,(fptr))&0x80000000)!=0)
#define isFpNonNegative(fptr) ((Dword(1,(fptr))&0x80000000)==0)
#define isFpPositive(fptr)    (isFpNonNegative(fptr)&&(!isFpZero(fptr)))
#define isFpNonPositive(fptr) (isFpNegative(fptr)||isFpZero(fptr))
#define doFpSignsDiffer(fptr1,fptr2) \
        (((Dword(1,(fptr1))^Dword(1,(fptr2)))&0x80000000)!=0)

#define FpSign(fptr)    (isFpNegative(fptr)?-1:(isFpZero(fptr)?0:1))
#define FpApplySign(fpx,fps) (isFpNegative(fps)?-(fpx):(fpx))
#define FpNegate(fptr)      {(Dword(1,(fptr))^=0x80000000);}
#define FpAbs(fptr)         {(Dword(1,(fptr))&=0x7fffffff);}
#define Fp2Exponent(fptr)   ((((LONG)(Dword(1,(fptr))&0x7ff00000))-0x3ff00000L)>>20)
#define Fp2Mantissa(fptr)   {Dword(1,(fptr)=(Dword(1,(fptr)&0x800fffff)+0x40000000;}
#define Fp2AbsMantissa(fptr){Dword(1,(fptr)=(Dword(1,(fptr)&0x000fffff)+0x40000000;}

#define GetFpERR(fptr)      (isFpNAN(fptr)?Dword(0,(fptr)):0)
#define SetFpERR(fptr,err)  {Dword(0,(fptr))=err;Dword(1,(fptr))=0xffff0000;}
#define SetFpPlusINF(fptr)  {Dword(0,(fptr))=0;Dword(1,(fptr))=0x7ff00000;}
#define SetFpMinusINF(fptr) {Dword(0,(fptr))=0;Dword(1,(fptr))=0xfff00000;}

// ===================================================================
// Globally accessible constants
// ===================================================================

// Not-a-Number
extern REAL ___NaN;
extern REAL ___PlusINF;
extern REAL ___MinusINF;

// Basic
extern REAL ___Zero;
extern REAL ___One;
extern REAL ___MaxReal;
extern REAL ___MinReal;
extern REAL ___EpsReal;
extern REAL ___SqrtEpsReal;
extern REAL ___LnMaxReal;
extern REAL ___LnMinReal;
extern REAL ___LnMinRealHalf;

// Algebraic
extern REAL ___Sqrt2;
extern REAL ___Sqrt3;
extern REAL ___Sqrt3_Div2;
extern REAL ___Sqrt5;
extern REAL ___Sqrt6;
extern REAL ___Sqrt7;
extern REAL ___Sqrt8;
extern REAL ___Sqrt10;
extern REAL ___GoldenRatio;  // ca 1.618 ...

// Transcendental
extern REAL ___Pi;
extern REAL ___PiHalf;
extern REAL ___TwoPi;
extern REAL ___SqrtPi;
extern REAL ___SqrtTwoPi;
extern REAL ___1DivSqrtTwoPi;
extern REAL ___Sqrt_4DivPi;
extern REAL ___Sqrt_PiDiv4Ln2;
extern REAL ___E;
extern REAL ___1DivE;
extern REAL ___Gamma;

// Natural logarithms
extern REAL ___Ln2;
extern REAL ___SqrtLn2;
extern REAL ___Ln10;
extern REAL ___LnPi;
extern REAL ___1dB_Power;
extern REAL ___1dB_Ampl;
extern REAL ___3dB_Power;
extern REAL ___3dB_Ampl;

// Decadic logarithms
extern REAL ___Log2;
extern REAL ___LogE;
extern REAL ___LogPi;

// Angular
extern REAL ___RadToDeg;
extern REAL ___DegToRad;
extern REAL ___MagicAngleRad;
extern REAL ___MagicAngleDeg;
extern REAL ___1DivSqrt2;  // also Sin45 and Cos45
extern REAL ___Sin30;
extern REAL ___Cos30;
extern REAL ___Sin60;
extern REAL ___Cos60;

// Probability
extern REAL ___ProbDev;       // Probasble deviation when stddev = 1
extern REAL ___1DivSqrt2ByPi; // Prob.density at x=0 when stddev = 1


// ===================================================================
// Exponent, mantissa, sign, rounding, truncating, tapering, ...
// ===================================================================

LONG lSign(REAL r);
REAL rSign(REAL r);

LONG lSignGE(REAL r);
REAL rSignGE(REAL r);

LONG lSignLE(REAL r);
REAL rSignLE(REAL r);

// When sign<0, returns -r, otherwise returns r
REAL FlipSign(REAL r,REAL sign);

// Returns rSign(sign)*fabs(r)
REAL AssignSign(REAL r,REAL sign);

// Returns the nearest integer, but in REAL format
REAL NearestRint(REAL x);

// Rounds symmetrically to the nearest integer multiple of eps.
REAL RoundToRint(REAL x,REAL eps=1.0);

// Exponent and/or mantissa extraction in any base (default is 10)
LONG ExponentAndMantissa(REAL f,REAL* m=NULL,DWORD base=10);
inline LONG Exponent(REAL f, DWORD base)
              {return ExponentAndMantissa(f,NULL,base);}
REAL Mantissa(REAL f, DWORD base);

LONG FloorFrac(REAL x,REAL &frac);
LONG SymmetricFloorFrac(REAL x,REAL &frac);
LONG RoundFrac(REAL x,REAL &frac);
LONG SymmetricRoundFrac(REAL x,REAL &frac);


// ===================================================================
// Cyclic, periodic and radial functions
// ===================================================================

// The decomposition x = quotient*mod + remainder,
REAL   RealModulus(REAL x,REAL mod,REAL* quotient = NULL);

// Radians-to-degrees conversion
inline REAL RadToDeg(REAL rad){return ___RadToDeg*rad;}

// Degrees-to-radians conversion
inline REAL DegToRad(REAL deg){return ___DegToRad*deg;}

// Safe version of atan
REAL   ArcTan(REAL x,REAL y);
inline REAL ArcTan(REAL y){return ArcTan(1.0,y);}

// Safe version of 1/atan
REAL   ArcCotan(REAL x,REAL y);
inline REAL ArcCotan(REAL y){return ArcCotan(1.0,y);}

// Safe azimuth calculation in radians
REAL   Azimuth(REAL x,REAL y);
inline REAL Azimuth(REAL y){return Azimuth(1.0,y);}

// Safe version of acos
REAL ArcCos(REAL x,REAL r);

// (x,y) amplitude/diagonal: square
inline REAL SquareAmplitude(REAL x,REAL y){return x*x+y*y;}

// (x,y) amplitude/diagonal: magnitude
inline REAL Amplitude(REAL x,REAL y){return sqrt(x*x+y*y);}


// ===================================================================
// Various peak functions
// ===================================================================

/*
// Centered Lorentzian with unit height and unit half-height width:
REAL LorentzHW(REAL x,REAL* imag);
// Centered Lorentzian with unit area and unit half-height width:
REAL LorentzAW(REAL x,REAL* imag);
*/

// ===================================================================
// Tapering, threshold, step, and interval-congruence functions
// ===================================================================

// Smoothed absolute value truncation from below
REAL SmoothLoBoundTaper(REAL val,REAL lowbound,REAL fac = 2.0);

// Smoothed absolute value truncation from above
REAL SmoothHiBoundTaper(REAL val,REAL uppbound,REAL fac = 2.0);

// Combines the previous two functions
REAL SmoothHiAndLoBoundsTaper
        (REAL x,REAL lobound,REAL hibound,REAL fac);

// This uses edge factors relative to the interval width
REAL SmoothIntervalTaper
        (REAL x,REAL lobound,REAL hibound,REAL relfac);

// Lorentzian soft-edged step function centered at zero.
REAL StepUpL(REAL x,REAL edge);

// As above but the G stands for "Gaussian" edge type.
REAL StepUpG(REAL x,REAL edge);

// A "soft" threshold giving 0 or 1 when x is much below or above xc.
REAL ThresholdL(REAL x,REAL xl,REAL xc,REAL xu,REAL sharpness = 3.0);

// As above, but with Gaussian edge shape, instead of Lorentzian.
REAL ThresholdG(REAL x,REAL xl,REAL xc,REAL xu,REAL sharpness = 3.0);

// Lorentzian threshold function with relative width of the edges.
REAL ThresholdRL(REAL x,REAL thr,REAL reledge,REAL sharpness = 3.0);

// As above, but with Gaussian edge shape, instead of Lorentzian.
REAL ThresholdRG(REAL x,REAL thr,REAL reledge,REAL sharpness = 3.0);

// Soft edge smoothing over an interval (rectangular function)
REAL EdgeRectL(REAL x,REAL lo,REAL hi,REAL loedg,REAL hiedg);

// As above, but with Gaussian edge shape, instead of Lorentzian.
REAL EdgeRectG(REAL x,REAL lo,REAL hi,REAL loedg,REAL hiedg);

// A soft-edged interval with Lorentzian edges, centered at 0.
REAL SymmetricBumpL(REAL x,REAL width,REAL edges);

// As above, but with Gaussian edge shape, instead of Lorentzian.
REAL SymmetricBumpG(REAL x,REAL width,REAL edges);

// A Lorentzian having different upper and lower half widths:
REAL BiLorentzian(REAL x,REAL xl,REAL xc,REAL xu);

// A Lorentzian with widths proportional to the center offset:
REAL BiLorentzian(REAL x,REAL xc,REAL relwidth);


// ===================================================================
// Other functions of a single variable
// ===================================================================

// x^n, with n a non-negative integer:
REAL Power(REAL x,DWORD n);
// x^n, with n an integer:
REAL Power(REAL x,LONG n);
// |x|^n, with n an integer:
REAL AbsPower(REAL x,LONG n);
// sign(x)*|x|^n, with n an integer:
REAL SignedPower(REAL x,LONG n);
// |x|^p, with p a REAL:
REAL Power(REAL x,REAL p);
// sign(x)*|x|^p, with p a REAL:
REAL SignedPower(REAL x,REAL p);

// Exponential-generating-function term factor:
REAL EgfFactor(REAL x,DWORD n);

// Complete elliptic integral of the first kind:
REAL EllipticK(REAL x);

// REAL version of the factorial of n:
REAL Factorial(DWORD n);

// Symmetric logarithm of x:
REAL SymLog(REAL x);
// Inverse of SymLog(x):
REAL SymLogInverse(REAL x);

// Symmetric decadic logarithm of x:
REAL SymLog10(REAL x);
// Inverse of SymLog10:
REAL SymLog10Inverse(REAL x);

// Sinc(x) = sin(x)/x
REAL Sinc(REAL x);


// ===================================================================
// Various functions of two variables
// ===================================================================

// Sqrt(a^2+b^2). Recommended also with LONG arguments.
REAL Pythag(REAL a,REAL b);

// Arithmetic sum
REAL AriSeriesSum(DWORD n,REAL first,REAL step);
REAL AriSeriesSum0(DWORD n,REAL first,REAL last);

// Geometric sum
REAL GeoSeriesSum(DWORD n,REAL first,REAL factor);
REAL GeoSeriesSum0(REAL first,REAL last,REAL factor);

// Binomial coefficient
REAL BinomialR(DWORD n,DWORD k);
REAL BinomialR(REAL alpha,DWORD k);

// Arithmetic-Geometric mean
REAL AGMean(REAL x,REAL y);

// Roots of quadratic equations
DWORD QuadraticEqRoots(REAL a,REAL b,REAL c,REAL* r1r,REAL* r2i);


// ===================================================================
// Interpolations/Extrapolations (noiseless)
// ===================================================================

// Interpolating tabulated data pairs: Newton
REAL NewtonInterpolation
    (REAL x,REALV xarr,REALV yarr,DWORD npairs,REALV dp,DWORD mode);
REAL NewtonInterpolation
    (REAL x,REALV xypairs,DWORD npairs,REALV dp,DWORD mode);

// Interpolating tabulated data pairs: Lagrange
REAL LagrangeInterpolation(REAL x,REALV xarr,REALV yarr,DWORD npairs);
REAL LagrangeInterpolation(REAL x, REALV xypairs, DWORD npairs);

// Interpolating tabulated data pairs: linear
REAL LinearInterpolation(REAL x,REALV xarr,REALV yarr,DWORD npairs);
REAL LinearInterpolation(REAL x,REALV xypairs,DWORD npairs);

// Interpolating tabulated data pairs: quadratic
REAL QuadraticInterpolation(REAL x,REALV xypairs,DWORD npairs);

// 3-point interpolations
// -------------------------------------------------------------------
// See http://dx.doi.org/10.3247/SL3Soft06.002

// Computes local zeroth, first and second derivative:
REAL P3InterpolateDer(REAL xl,REAL xc,REAL xu,
                      REAL yl,REAL yc,REAL yu,REAL &d0,REAL &d1);

// As above, but for equidistant points, xl = xc-d and xu = xc+d:
REAL P3InterpolateDer_e
        (REAL d,REAL yl,REAL yc,REAL yu,REAL &d0,REAL &d1);

// As above, but with the distance d being 1.0:
REAL P3InterpolateDer_1(REAL yl,REAL yc,REAL yu,REAL &d0,REAL &d1);

// Calculates y(x) using the values of xc,d0,d1 and d2
REAL P3Interpolate(REAL x, REAL xc,REAL d0,REAL d1,REAL d2);

// Estimates and returns y(x); full computation
REAL P3Interpolate_f
        (REAL x,REAL xl,REAL xc,REAL xu,REAL yl,REAL yc,REAL yu);

// Estimates and returns y(x) when xl = xc-d and xu = xc+d
REAL P3Interpolate_e(REAL x,REAL xc,REAL d,REAL yl,REAL yc,REAL yu);

// As above, but with the distance d being 1.0:
REAL P3Interpolate_1(REAL x,REAL xc,REAL yl,REAL yc,REAL yu);

// 3-point interpolation (quadratic); compact code for general case
REAL Q3PInterpolation(REAL x,REAL xl,REAL xc,REAL xu,
                 REAL yl,REAL yc,REAL yu);

// 3-point extremum estimate; compact code for general case
REAL Q3PExtremum(REAL xl,REAL xc,REAL xu,
                 REAL yl,REAL yc,REAL yu,REAL &xmin,REAL &ymin);

//--------------------------------------------------------------------
#endif



